
import functools
from types import MethodType
from typing import Coroutine, Dict, Tuple, Union

from brownie.network.contract import ContractTx, _get_tx
from dank_mids.brownie_patch.account import _patch_account
from web3 import Web3


def _patch_tx(call: ContractTx, w3: Web3) -> None:

    @functools.wraps(call)
    async def coroutine(
        self,
        *args: Tuple,
        block_identifier: Union[int, str, bytes] = None,
        override: Dict = None
    ) -> Coroutine:

        args, tx = _get_tx(self._owner, args)
        if not tx["from"]:
            raise AttributeError(
                "Final argument must be a dict of transaction parameters that "
                "includes a `from` field specifying the sender of the transaction"
            )
        
        if not hasattr(tx["from"], 'transfer_coro'):
           tx["from"] = _patch_account(tx["from"], w3)

        return await tx["from"].transfer_coro(
            self._address,
            tx["value"],
            gas_limit=tx["gas"],
            gas_buffer=tx.get("gas_buffer"),
            gas_price=tx.get("gas_price"),
            max_fee=tx.get("max_fee"),
            priority_fee=tx.get("priority_fee"),
            nonce=tx["nonce"],
            required_confs=tx["required_confs"],
            data=self.encode_input(*args),
            allow_revert=tx["allow_revert"],
        )

    call.coroutine = MethodType(coroutine, call)




import asyncio
import threading
from collections import defaultdict
from functools import lru_cache
from time import time
from typing import Any, Dict, List, Literal, Optional, Union

import multicall
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from multicall.utils import gather, run_in_subprocess
from web3 import Web3
from web3.providers import HTTPProvider
from web3.providers.async_base import AsyncBaseProvider
from web3.types import RPCEndpoint, RPCResponse

from dank_mids.constants import (BAD_HEXES, GAS_LIMIT, LOOP_INTERVAL,
                                 OVERRIDE_CODE)
from dank_mids.loggers import (demo_logger, main_logger, sort_lazy_logger,
                               sort_logger)

instances: List["DankMiddlewareController"] = []

def reattempt_call_and_return_exception(target: str, calldata: bytes, block: str, web3: Web3) -> Exception:
    """ NOTE: This runs synchronously in a subprocess in order to bypass Dank Middleware without blocking the event loop. """
    try:
        return web3.eth.call({"to": target, "data": calldata}, block)
    except Exception as e:
        return e

def _err_msg(e: Exception) -> str:
    """ Extract an error message from `e` to use in a spoof rpc response. """
    if isinstance(e.args[0], str):
        err_msg = f"DankMidsError: {type(e)} {e.args}"
    elif "message" in e.args[0]:
        err_msg = e.args[0]["message"]
    elif "error" in e.args[0] and "message" in e.args[0]["error"]:
        err_msg = e.args[0]["error"]["message"]
    return err_msg


class DankMiddlewareController:
    def __init__(self, w3: Web3) -> None:
        assert w3.eth.is_async and isinstance(w3.provider, AsyncBaseProvider), "Dank Middleware can only be applied to an asycnhronous Web3 instance."
        self.w3 = w3
        self.sync_w3 = Web3(provider = HTTPProvider(self.w3.provider.endpoint_uri))
        # Can't pickle middlewares to send to process executor
        self.sync_w3.middleware_onion.clear()
        self.DO_NOT_BATCH = set()
        self.pending_calls = defaultdict(dict)
        self.completed_calls = defaultdict(dict)
        self.batcher = multicall.multicall.batcher
        self.process_executor = multicall.utils.process_pool_executor
        self._bid = 0   #      batch id
        self._mid = 0   #  multicall id
        self._cid = 0   #       call id
        self._bid_lock = threading.Lock()
        self._mid_lock = threading.Lock()
        self._cid_lock = threading.Lock()
        self._last_seen_cid = 0
        self._initializing = False
        self._is_configured = False
        self._is_executing = False
        self._pools_closed = False
        self._checkpoint = time()
        self._instance = len(instances)
        instances.append(self)
    
    def __repr__(self) -> str:
        return f"<DankMiddlewareController {self._instance}>"

    async def __call__(self, params: Any) -> RPCResponse:
        if not self._is_configured:
            await self._setup()
        block = params[1]
        cid = await self.add_to_queue(params)
        return await self.await_response(block,cid)
    
    @property
    def next_bid(self) -> int:
        return self._increment('bid')
    
    @property
    def next_mid(self) -> int:
        return self._increment('mid')
    
    @property
    def next_cid(self) -> int:
        return self._increment('cid')
    
    @sort_lazy_logger
    def should_batch(self, method: RPCEndpoint, params: Any) -> bool:
        if method != 'eth_call':
            sort_logger.debug(f"bypassed, method is {method}")
            return False
        elif params[0]['to'] in self.DO_NOT_BATCH:
            sort_logger.debug(f"bypassed, target is in `DO_NOT_BATCH`")
            return False
        return True
    
    async def add_to_queue(self, params: Any) -> int:
        cid = self.next_cid
        block = params[1]
        while self._pools_closed:
            await asyncio.sleep(0)
        self.pending_calls[block][cid] = params
        demo_logger.info(f'added to queue (cid: {cid})')
        return cid
    
    async def await_response(self, block: str, cid: int) -> RPCResponse:
        while cid not in self.completed_calls[block]:
            assert cid in self.pending_calls[block], f"Something went wrong, call{cid} is missing from `pending_calls`."
            if self.queue_is_ready() and self.loop_is_ready():
                await self.execute_multicall()
            await asyncio.sleep(LOOP_INTERVAL)
            self._last_seen_cid = self._cid
        return self.fetch_response(block,cid)

    def queue_is_ready(self) -> bool:
        if not self._is_executing and self._cid == self._last_seen_cid and len(self.pending_calls):
            return True
        return False
    
    def loop_is_ready(self) -> bool:
        return time() - self._checkpoint > LOOP_INTERVAL
    
    def fetch_response(self, block: str, cid: int) -> RPCResponse:
        response = self.completed_calls[block].pop(cid)
        if not self.completed_calls[block]:
            self.completed_calls.pop(block)
        return response
    
    async def execute_multicall(self) -> None:
        i = 0
        while self._cid_lock.locked():
            if i // 50 == int(i // 50):
                print('lock is locked')
            await asyncio.sleep(.1)
        if self._is_executing:
            return
        demo_logger.info(f'executing multicall (current cid: {self._cid})')
        self._is_executing = True
        self._pools_closed = True
        with self._cid_lock:
            pending_calls = list(self.pending_calls.items())
        self._pools_closed = False
        await gather([self.process_block(block, calls) for block, calls in pending_calls])
        self._is_executing = False
        demo_logger.info('multicall complete')
    
    async def process_block(self, block: str, calls: Dict[int,List]) -> None:
        calls = [[cid, [params[0]['to'], HexBytes(params[0]['data'])]] for cid, params in calls.items()]
        demo_logger.info(f'executing {len(calls)} calls for block {block}')
        batches = self.batcher.batch_calls(calls, self.batcher.step)
        batches = await gather([self.process_batch(batch,block) for batch in batches])

    async def process_batch(self, batch: List, block: str, bid: Optional[int] = None) -> None:
        if bid is None:
            bid = self.next_bid
        mid = self.next_mid
        demo_logger.info(f'tryBlockAndAggregate {mid} for batch {bid} starting')
        batch = dict(batch)
        cids = list(batch.keys())
        inputs = list(batch.values())
        try:
            _, _, response = await self._multicall_for_block(block).coroutine([False, inputs])
            demo_logger.info(f'tryBlockAndAggregate {mid} for batch {bid} complete')
            await gather([
                self.spoof_response(cid, block, params, data)
                for cid, params, (_, data) in zip(cids, inputs, response)
            ])

        except Exception as e:
            if len(inputs) == 1:
                await self.spoof_response(cids[0], block, inputs[0], e)
                return
            elif "out of gas" in str(e):
                # TODO Remember which contracts/calls are gas guzzlers
                main_logger.debug('out of gas. cut in half, trying again')
            elif any(err in str(e).lower() for err in ["connection reset by peer","request entity too large","server disconnected"]):
                main_logger.debug('dank too loud, trying again')
                old_step = self.batcher.step
                self.batcher.step = round(len(inputs) * 0.99) if len(inputs) >= 100 else len(inputs) - 1
                main_logger.warning(f'Multicall batch size reduced from {old_step} to {self.batcher.step}. The failed batch had {len(inputs)} calls.')
            else:
                #raise
                print(f"unexpected exception: {type(e)} {str(e)}")

            batches = [[cid, input] for cid, input in zip(cids, inputs)]
            halfpoint = len(batches) // 2
            await gather([
                self.process_batch(batches[:halfpoint], block, str(bid)+"_0"),
                self.process_batch(batches[halfpoint:], block, str(bid)+"_1"),
            ])
            demo_logger.info(f'tryBlockAndAggregate {mid} for batch {bid} complete')
    
    @lru_cache(maxsize=None)
    def _multicall_for_block(self, block: str) -> multicall.Call:
        return multicall.Call(
            self.MULTICALL2,
            "tryBlockAndAggregate(bool,(address,bytes)[])(uint256,uint256,(bool,bytes)[])",
            returns=None,
            block_id=block,
            gas_limit=GAS_LIMIT,
            state_override_code=OVERRIDE_CODE,
        )
    
    async def spoof_response(self, cid: int, block: str, params: List, data: Optional[Union[str,Exception]]) -> RPCResponse:
        """ Creates a spoof rpc response for call # `cid` using data returned from multicall2. """
        if (
            # If multicall failed, try single call to get detailed error.
            data is None
            # Or, if we got a known "bad" response from the multicall, try single call.
            # Could be that target contract does not support multicall2.
            or (isinstance(data, bytes) and HexBytes(data).hex() in BAD_HEXES)
        ):
            target, calldata = params
            data = await run_in_subprocess(reattempt_call_and_return_exception, target, calldata, block, self.sync_w3)
            # We were able to get a usable response from single call.
            # Add contract to DO_NOT_BATCH list
            if not isinstance(data, Exception):
                self.DO_NOT_BATCH.add(target) 
        if isinstance(data, Exception):
            spoof = {"error": {'code': -32000, 'message': _err_msg(data)}}
        else:
            spoof = {"result": HexBytes(data).hex()}
        spoof.update({"id": cid, "jsonrpc": "dank_mids"})
        main_logger.debug(f"spoof: {spoof}")
        self.completed_calls[block][cid] = spoof

        # Pop the call from pending calls
        # TODO figure out why cids can be missing from pending_calls when they haven't been popped yet
        if cid in self.pending_calls[block]:
            self.pending_calls[block].pop(cid)

        # Pop the block from pending calls if empty
        if not self.pending_calls[block]:
            self.pending_calls.pop(block)
        return data

    async def _setup(self) -> None:
        if self._initializing:
            while self._initializing:
                await asyncio.sleep(0)
            return
        self._initializing = True
        print('Dank Middleware initializing... Strap on your rocket boots...')
        chain_id = await self.w3.eth.chain_id
        MULTICALL = multicall.constants.MULTICALL_ADDRESSES.get(chain_id,None)
        self.MULTICALL2 = multicall.constants.MULTICALL2_ADDRESSES.get(chain_id,None)
        self.DO_NOT_BATCH.update(to_checksum_address(address) for address in [MULTICALL,self.MULTICALL2] if address)
        self._is_configured = True
        self._initializing = False
    
    def _increment(self, id: Literal["bid","mid","cid"]) -> int:
        attr = f"_{id}"
        with getattr(self, f"{attr}_lock"):
            new = getattr(self, attr) + 1
            setattr(self, attr, new)
            return new




from brownie import Contract
from dank_mids.brownie_patch import patch_contract
from dank_mids.brownie_patch.contractcall import _patch_call
from dank_mids.brownie_patch.tx import _patch_tx
from multicall.utils import await_awaitable

from tests.fixtures import dank_w3

def test_patch_call():
    # must use from_explorer for gh testing workflow
    weth = Contract.from_explorer('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2')
    _patch_call(weth.totalSupply, dank_w3)
    assert hasattr(weth.totalSupply, 'coroutine')
    assert await_awaitable(weth.totalSupply.coroutine(block_identifier=13_000_000)) == 6620041514474872981393155

# TODO implement this test once _patch_tx is implemented
#def test_patch_tx():
    # must use from_explorer for gh testing workflow
#    uni_v3_quoter = Contract.from_explorer('0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6')
#    _patch_tx(uni_v3_quoter.quoteExactInput, dank_w3)
#    assert hasattr(uni_v3_quoter.quoteExactInput, 'coroutine')
#    assert await_awaitable(uni_v3_quoter.quoteExactInput.coroutine(block_identifier=13_000_000)) == 0

def test_patch_contract():
    # ContractCall
    # must use from_explorer for gh testing workflow
    weth = patch_contract(Contract.from_explorer('0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2'), dank_w3)
    assert hasattr(weth.totalSupply, 'coroutine')
    assert await_awaitable(weth.totalSupply.coroutine(block_identifier=13_000_000)) == 6620041514474872981393155

    # ContractTx
    # TODO implement this part of the test once _patch_tx is implemented
    # must use from_explorer for gh testing workflow
    #uni_v3_quoter = patch_contract(Contract.from_explorer('0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'), dank_w3)
    #assert hasattr(uni_v3_quoter.quoteExactInput, 'coroutine')
    #assert await_awaitable(uni_v3_quoter.quoteExactInput.coroutine(block_identifier=13_000_000)) == 0



from typing import Any, Callable, Coroutine

from web3.types import RPCEndpoint, RPCResponse

# This was introduced in web3 v5.28.0 but we like loose deps here
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]



import functools
from types import MethodType
from typing import Coroutine, Dict, Tuple, Union

import eth_abi
from brownie.exceptions import VirtualMachineError
from brownie.network.contract import ContractCall, _get_tx
from brownie.project.compiler.solidity import SOLIDITY_ERROR_CODES
from hexbytes import HexBytes
from web3 import Web3


def _patch_call(call: ContractCall, w3: Web3) -> None:
    @functools.wraps(call)
    async def coroutine(
        self,
        *args: Tuple,
        block_identifier: Union[int, str, bytes] = None,
        override: Dict = None
    ) -> Coroutine:

        args, tx = _get_tx(self._owner, args)
        if tx["from"]:
            tx["from"] = str(tx["from"])
        del tx["required_confs"]
        tx.update({"to": self._address, "data": self.encode_input(*args)})

        try:
            data = await w3.eth.call({k: v for k, v in tx.items() if v}, block_identifier, override)
        except ValueError as e:
            raise VirtualMachineError(e) from None

        selector = HexBytes(data)[:4].hex()

        if selector == "0x08c379a0":
            revert_str = eth_abi.decode_abi(["string"], HexBytes(data)[4:])[0]
            raise ValueError(f"Call reverted: {revert_str}")
        elif selector == "0x4e487b71":
            error_code = int(HexBytes(data)[4:].hex(), 16)
            if error_code in SOLIDITY_ERROR_CODES:
                revert_str = SOLIDITY_ERROR_CODES[error_code]
            else:
                revert_str = f"Panic (error code: {error_code})"
            raise ValueError(f"Call reverted: {revert_str}")
        if self.abi["outputs"] and not data:
            raise ValueError("No data was returned - the call likely reverted")
        return self.decode_output(data)

    call.coroutine = MethodType(coroutine, call)




import sys
import threading
import time
from types import MethodType
from typing import Any, Dict, Optional, Tuple

from brownie._config import CONFIG
from brownie.convert import Wei
from brownie.exceptions import VirtualMachineError
from brownie.network.account import Account
from brownie.network.rpc import Rpc
from brownie.network.state import Chain
from brownie.network.transaction import TransactionReceipt
from web3 import Web3

rpc = Rpc()

def _patch_account(account: Account, w3: Web3) -> Account:
    async def transfer_coro(
        self,
        to: Account = None,
        amount: int = 0,
        gas_limit: Optional[int] = None,
        gas_buffer: Optional[float] = None,
        gas_price: Optional[int] = None,
        max_fee: Optional[int] = None,
        priority_fee: Optional[int] = None,
        data: str = None,
        nonce: Optional[int] = None,
        required_confs: int = 1,
        allow_revert: bool = None,
        silent: bool = None,
    ) -> TransactionReceipt:
        """
        Broadcast a transaction from this account.
        Kwargs:
            to: Account instance or address string to transfer to.
            amount: Amount of ether to send, in wei.
            gas_limit: Gas limit of the transaction.
            gas_buffer: Multiplier to apply to gas limit.
            gas_price: Gas price of legacy transaction.
            max_fee: Max fee per gas of dynamic fee transaction.
            priority_fee: Max priority fee per gas of dynamic fee transaction.
            nonce: Nonce to use for the transaction.
            data: Hexstring of data to include in transaction.
            silent: Toggles console verbosity.
        Returns:
            TransactionReceipt object
        """

        receipt, exc = await self._make_transaction_coro(
            to,
            amount,
            gas_limit,
            gas_buffer,
            gas_price,
            max_fee,
            priority_fee,
            data or "",
            nonce,
            "",
            required_confs,
            allow_revert,
            silent,
        )

        if rpc.is_active():
            undo_thread = threading.Thread(
                target=Chain()._add_to_undo_buffer,
                args=(
                    receipt,
                    self.transfer,
                    [],
                    {
                        "to": to,
                        "amount": amount,
                        "gas_limit": gas_limit,
                        "gas_buffer": gas_buffer,
                        "gas_price": gas_price,
                        "max_fee": max_fee,
                        "priority_fee": priority_fee,
                        "data": data,
                    },
                ),
                daemon=True,
            )
            undo_thread.start()

        receipt._raise_if_reverted(exc)
        return receipt
    
    async def _make_transaction_coro(
        self,
        to: Optional["Account"],
        amount: int,
        gas_limit: Optional[int],
        gas_buffer: Optional[float],
        gas_price: Optional[int],
        max_fee: Optional[int],
        priority_fee: Optional[int],
        data: str,
        nonce: Optional[int],
        fn_name: str,
        required_confs: int,
        allow_revert: Optional[bool],
        silent: Optional[bool],
    ) -> Tuple[TransactionReceipt, Optional[Exception]]:
        # shared logic for `transfer` and `deploy`
        if gas_limit and gas_buffer:
            raise ValueError("Cannot set gas_limit and gas_buffer together")
        if silent is None:
            silent = bool(CONFIG.mode == "test" or CONFIG.argv["silent"])

        if gas_price is None:
            # if gas price is not explicitly set, load the default max fee and priority fee
            if max_fee is None:
                max_fee = CONFIG.active_network["settings"]["max_fee"] or None
            if priority_fee is None:
                priority_fee = CONFIG.active_network["settings"]["priority_fee"] or None

        if priority_fee == "auto":
            priority_fee = Chain().priority_fee

        try:
            # if max fee and priority fee are not set, use gas price
            if max_fee is None and priority_fee is None:
                gas_price, gas_strategy, gas_iter = self._gas_price(gas_price)
            else:
                gas_strategy, gas_iter = None, None
            gas_limit = Wei(gas_limit) or self._gas_limit(
                to, amount, gas_price or max_fee, gas_buffer, data
            )
        except ValueError as e:
            raise VirtualMachineError(e) from None

        with self._lock:
            # we use a lock here to prevent nonce issues when sending many tx's at once
            tx = {
                "from": self.address,
                "value": Wei(amount),
                "nonce": nonce if nonce is not None else self._pending_nonce(),
                "gas": web3.toHex(gas_limit),
                "data": HexBytes(data),
            }
            if to:
                tx["to"] = to_address(str(to))
            tx = _apply_fee_to_tx(tx, gas_price, max_fee, priority_fee)
            txid = None
            while True:
                try:
                    response = await self._transact(tx, allow_revert)  # type: ignore
                    exc, revert_data = None, None
                    if txid is None:
                        txid = HexBytes(response).hex()
                        if not silent:
                            print(f"\rTransaction sent: {color('bright blue')}{txid}{color}")
                except ValueError as e:
                    if txid is None:
                        exc = VirtualMachineError(e)
                        if not hasattr(exc, "txid"):
                            raise exc from None
                        txid = exc.txid
                        print(f"\rTransaction sent: {color('bright blue')}{txid}{color}")
                        revert_data = (exc.revert_msg, exc.pc, exc.revert_type)
                try:
                    receipt = TransactionReceipt(
                        txid,
                        self,
                        silent=silent,
                        required_confs=required_confs,
                        is_blocking=False,
                        name=fn_name,
                        revert_data=revert_data,
                    )  # type: ignore
                    break
                except (TransactionNotFound, ValueError):
                    if not silent:
                        sys.stdout.write(f"  Awaiting transaction in the mempool... {_marker[0]}\r")
                        sys.stdout.flush()
                        _marker.rotate(1)
                    time.sleep(1)

        receipt = self._await_confirmation(receipt, required_confs, gas_strategy, gas_iter)
        if receipt.status != 1 and exc is None:
            error_data = {
                "message": (
                    f"VM Exception while processing transaction: revert {receipt.revert_msg}"
                ),
                "code": -32000,
                "data": {
                    receipt.txid: {
                        "error": "revert",
                        "program_counter": receipt._revert_pc,
                        "return": receipt.return_value,
                        "reason": receipt.revert_msg,
                    },
                },
            }
            exc = VirtualMachineError(ValueError(error_data))

        return receipt, exc
    
    async def _transact_coro(self, tx: Dict, allow_revert: bool) -> Any:
        if allow_revert is None:
            allow_revert = bool(CONFIG.network_type == "development")
        if not allow_revert:
            await self._check_for_revert_coro(tx)
        return await w3.eth.send_transaction(tx)
    
    async def _check_for_revert_coro(self, tx: Dict) -> None:
        try:
            # remove gas price related values to avoid issues post-EIP1559
            # https://github.com/ethereum/go-ethereum/pull/23027
            skip_keys = {"gasPrice", "maxFeePerGas", "maxPriorityFeePerGas"}
            await w3.eth.call({k: v for k, v in tx.items() if k not in skip_keys and v})
        except ValueError as exc:
            msg = exc.args[0]["message"] if isinstance(exc.args[0], dict) else str(exc)
            raise ValueError(
                f"Execution reverted during call: '{msg}'. This transaction will likely revert. "
                "If you wish to broadcast, include `allow_revert:True` as a transaction parameter.",
            ) from None
    
    account.transfer_coro = MethodType(transfer_coro, account)
    account._make_transaction_coro = MethodType(_make_transaction_coro, account)
    account._transact_coro = MethodType(_transact_coro, account)
    account._check_for_revert_coro = MethodType(_check_for_revert_coro, account)
    return account
